import json
from pathlib import Path
from typing import Any, Dict, Union

from . import markers

__all__ = ["matplotlib_style", "markers"]

matplotlib_style: Dict[str, Dict[str, Union[str, Any]]] = json.loads(
    (Path(__file__).parent / "matplotlib_style.json").read_text()
)

for _, value in matplotlib_style.items():
    if "marker" in value.keys():
        value["marker"] = getattr(markers, value["marker"])



import json
import logging
import re
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional

from appdirs import user_cache_dir
from bs4 import BeautifulSoup

from ...utils.cache import CacheResults
from ..requests import session

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

__all__ = ["features", "taglist"]  # noqa: F822

_log = logging.getLogger(__name__)


class TaglistType(TypedDict):
    features: List[str]
    taglist: Dict[str, List[str]]


JSONType = Any


def _hash_request(*args, **kwargs) -> str:
    return "taglist.json"


def _write_json(json_: JSONType, cache_file) -> None:  # coverage: ignore
    _log.info(f"Writing cache file {cache_file}")
    cache_file.write_text(json.dumps(json_, indent=2))


def _read_json(cache_file: Path) -> Optional[JSONType]:
    _log.info(f"Reading cache file {cache_file}")
    json_ = json.loads(cache_file.read_text())
    return json_


@CacheResults(
    cache_dir=Path(user_cache_dir("cartes")) / "osm",
    hashing=_hash_request,
    reader=_read_json,
    writer=_write_json,
)
def _feat_tags() -> TaglistType:

    c = session.get("https://wiki.openstreetmap.org/wiki/Map_Features")
    c.raise_for_status()
    e = BeautifulSoup(c.content)

    features = sorted(
        set(
            a.attrs["title"].split(":")[1].replace(" ", "_")
            for a in e.find_all("a", attrs={"href": re.compile("/wiki/Key:")})
        )
    )

    taglist = dict(
        (
            d.attrs["data-taginfo-taglist-tags"].split("=")[0],
            d.attrs["data-taginfo-taglist-tags"].split("=")[1].split(","),
        )
        for d in e.find_all("div", attrs={"class": "taglist"})
    )

    return dict(features=features, taglist=taglist)


def __getattr__(name: str) -> Any:
    if name in ["features", "taglist"]:
        return _feat_tags().get(name)
    raise AttributeError(f"No attribute '{name}'")



import itertools
from operator import itemgetter

from shapely.geometry.multilinestring import MultiLineString
from shapely.ops import linemerge, unary_union

from .. import Overpass
from ..core import Relation


class Waterway(Relation):
    """A class to parse waterway=* relations.

    The purpose of waterway relations is to have an object for each river.

    Reference: https://wiki.openstreetmap.org/wiki/Relation:waterway

    Tags:
      - type (waterway)
      - waterway [river, stream, canal, drain, ditch]
      - name (*)
      - destination ?
      - ref ?
      - ref:sandre ?
      - ref:fgkz ?
      - ...

    Relation members:
      - None or main_stream 1+
      - side_stream 0+
      - spring ?
      - tributary 0+

    """

    def __init__(self, json):

        super().__init__(json)

        self.parent: Overpass = json["_parent"]
        parsed_keys = dict(
            (elt["ref"], elt) for elt in self.parent.all_members[json["id_"]]
        )

        parts = dict(
            (
                role,
                unary_union(
                    list(
                        elt["geometry"]
                        for elt in it
                        if elt.get("geometry", None)
                    )
                ),
            )
            for role, it in itertools.groupby(
                parsed_keys.values(), key=itemgetter("role")
            )
        )
        main_stream = parts.get("main_stream", parts.get(""))
        elements = [
            linemerge(main_stream)
            if isinstance(main_stream, MultiLineString)
            else main_stream
        ]
        side_stream = parts.get("side_stream", None)
        if side_stream is not None:
            elements.append(
                linemerge(side_stream)
                if isinstance(side_stream, MultiLineString)
                else side_stream
            )

        list_ = list(elt for elt in elements if elt is not None)
        if len(list_) == 0:
            self.json["geometry"] = self.shape = None
        self.json["geometry"] = self.shape = unary_union(list_)



# type: ignore

"""
This module provides facilities for producing CRS projections.

All projections defined here are simple aliases of official EPSG definitions.
Definitions are parsed and dispatched to inherit from the proper Cartopy CRS
classes. Sometimes, there seems to be errors in the bounding boxes so there
are adjustments below, overwriting the boundind boxes as class attributes.

>>> Lambert93().proj4_init
'+ellps=GRS80 +proj=lcc +lon_0=3 +lat_0=46.5 +x_0=700000 ...'

>>> LCCEurope().bbox
{'south_latitude': 24.6...}

>>> LCCEurope().proj4_params
{'ellps': 'GRS80', 'proj': 'lcc', ...}

"""

# fmt: off
import warnings

# All projections first come from the regular Cartopy CRS module
from cartopy.crs import (  # noqa: F401, F403
    Globe,
    InterruptedGoodeHomolosine,
    LambertCylindrical,
    Miller,
    Projection,
    Robinson,
    RotatedPole,
    Sinusoidal,
)

from . import meta
from .info import valid_crs  # noqa: F401

# These ones can be unpacked in altair!
from .vega_params import (  # noqa: F401, F403
    OSGB,
    OSNI,
    UTM,
    AlbersEqualArea,
    AzimuthalEquidistant,
    EqualEarth,
    EquidistantConic,
    EuroPP,
    Gnomonic,
    LambertAzimuthalEqualArea,
    LambertConformal,
    Mercator,
    Mollweide,
    Orthographic,
    PlateCarree,
    Stereographic,
    TransverseMercator,
)

# fmt: on

# Silence a fiona warning
warnings.simplefilter(action="ignore", category=UserWarning)


# Official projection in Switzerland
CH1903p = meta.EPSG_2056

# Official projection in France
Lambert93 = meta.EPSG_2154


# Official projection in Sweden (adjusted to the whole country)
# This record seems to be centered on Stockholm agglomeration
class ST74(meta.EPSG_3152):
    bbox = {
        "east_longitude": 24.17,
        "north_latitude": 69.07,
        "south_latitude": 54.96,
        "west_longitude": 10.03,
    }


# Official projection in Japan (adjusted to the whole country)
class JGD2000(meta.EPSG_2451):
    bbox = {
        "east_longitude": 157.65,
        "north_latitude": 46.05,
        "south_latitude": 17.09,
        "west_longitude": 122.38,
    }


# Official projection in the Netherlands
Amersfoort = meta.EPSG_28992

# Lambert projection of Europe
LCCEurope = meta.EPSG_3034

# Official projection in Australia
GeoscienceAustraliaLambert = meta.EPSG_3112


# Official projection in Canada
# Bounding box adjusted because of strong distortions
class StatisticsCanadaLambert(meta.EPSG_3348):
    bbox = {
        "east_longitude": -50.0,
        "north_latitude": 85.0,
        "south_latitude": 32.0,
        "west_longitude": -131.1,
    }


# Official projection in Singapore
SingaporeTM = meta.EPSG_3414

# Official projection in New Zealand
NZGD2000 = meta.EPSG_3851

# Official projection in Korea
Korea2000 = meta.EPSG_5179


# Official projection in Latin America
class SIRGAS2000(meta.EPSG_5641):
    bbox = {
        "east_longitude": -25.28,
        "north_latitude": 32.72,
        "south_latitude": -59.87,
        "west_longitude": -122.19,
    }


# Official projection in Germany
class GaussKruger(meta.EPSG_5680):
    bbox = {
        "east_longitude": 15.0,
        "north_latitude": 55.09,
        "south_latitude": 47.27,
        "west_longitude": 5.87,
    }


GaussKrueger = GaussKruger

# Official projection of the United States
AlbersUSA = meta.EPSG_6350

# Official projection in Italy
RDN2008 = meta.EPSG_6875

# Official projection in South Africa
Hartebeesthoek94 = meta.EPSG_9221


def __getattr__(name: str) -> Projection:  # noqa: F405
    return getattr(meta, name)



from typing import Dict

import geopandas as gpd

from pyproj import Proj, Transformer
from shapely.geometry import MultiPolygon, Polygon, base, polygon
from shapely.ops import polygonize, transform


def reorient(shape: base.BaseGeometry, orientation=-1) -> base.BaseGeometry:
    if isinstance(shape, Polygon):
        return polygon.orient(shape, orientation)
    if isinstance(shape, MultiPolygon):
        return MultiPolygon(reorient(p) for p in shape.geoms)
    return shape


def simplify(
    shape: base.BaseGeometry, resolution: float, bounds: Dict[str, float]
) -> base.BaseGeometry:
    """Geometric shape simplification.

    Projects a given geometry to an equivalent projection before calling the
    usual Douglas-Peucker simplification.

    Reference: https://shapely.readthedocs.io/en/stable/manual.html#object.simplify
    """
    proj = Proj(
        proj="aea",  # equivalent projection
        lat_1=bounds["minlat"],
        lat_2=bounds["maxlat"],
        lat_0=(bounds["minlat"] + bounds["maxlat"]) / 2,
        lon_0=(bounds["minlon"] + bounds["maxlon"]) / 2,
    )
    transformer_fwd = Transformer.from_proj(
        Proj("EPSG:4326"), proj, always_xy=True
    )
    transformer_back = Transformer.from_proj(
        proj, Proj("EPSG:4326"), always_xy=True
    )
    return transform(
        transformer_back.transform,
        transform(transformer_fwd.transform, shape).simplify(resolution),
    )


def fix_polygon(elt: base.BaseGeometry) -> MultiPolygon:
    """Tentative to fix an invalid (multi)polygon."""
    return MultiPolygon(list(polygonize(elt)))


def fix_geodataframe(gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Fix (in place) geometries in a GeoDataFrame."""
    gdf = gdf.copy()
    if ~gdf.is_valid.sum():
        fixed = gdf.loc[~gdf.is_valid, "geometry"].apply(fix_polygon)
        gdf.loc[~gdf.is_valid, "geometry"] = fixed
    return gdf



from typing import Iterator, List, Tuple

import pandas as pd


class HTMLMixin:
    """Base class for HTML repr mixins.

    It may look vain, but it terminates the super()._repr_html_() cascade.
    """

    def _repr_html_(self) -> str:
        return ""


class HTMLTitleMixin(HTMLMixin):
    """A Mixin class to have the name of the class in the HTML repr."""

    def _repr_html_(self):
        return f"<h2>{type(self).__name__}</h2>" + super()._repr_html_()


class HTMLAttrMixin(HTMLMixin):
    """A Mixin class to build a table of attributes in the HTML repr.

    Classes using this Mixin should override the html_attr_list as an attribute
    or a property.
    """

    html_attr_list: List[str] = []

    def _expand_view(self) -> Iterator[Tuple[str, str]]:
        for key in self.html_attr_list:
            value = getattr(self, key, None)
            if value is None:
                continue
            if isinstance(value, list):
                for idx, v_ in enumerate(value):
                    yield key + f"_{idx}", v_
            elif isinstance(value, dict):
                yield from value.items()
            else:
                yield key, value

    def _attr_view(self) -> pd.DataFrame:
        return pd.DataFrame.from_records(  # one-trick poney
            [dict(elt for elt in self._expand_view())]
        ).T.rename(columns={0: ""})

    def _repr_html_(self):
        return (
            "<div style='float: left;'>"
            + self._attr_view()._repr_html_()
            + "</div>"
            + super()._repr_html_()
        )


class _HBox(object):
    def __init__(self, *args):
        self.elts = args

    def _repr_html_(self):
        return "".join(
            f"""
    <div style='
        margin: 1ex;
        min-width: 250px;
        max-width: 300px;
        display: inline-block;
        vertical-align: top;'>
        {elt._repr_html_()}
    </div>
    """
            for elt in self.elts
        )

    def __or__(self, other) -> "_HBox":
        if isinstance(other, _HBox):
            return _HBox(*self.elts, *other.elts)
        else:
            return _HBox(*self.elts, other)

    def __ror__(self, other) -> "_HBox":
        if isinstance(other, _HBox):
            return _HBox(*other.elts, *self.elts)
        else:
            return _HBox(other, *self.elts)


class HBoxMixin(object):
    """Enables a | operator for placing representations next to each other."""

    def __or__(self, other) -> _HBox:
        if isinstance(other, _HBox):
            return _HBox(self, *other.elts)
        else:
            return _HBox(self, other)



from typing import List, Optional, Type, TypeVar, Union

from shapely.geometry import mapping, shape

from ..core import GeoObject
from ..utils.descriptors import OrientedShape
from ..utils.mixins import HBoxMixin, HTMLAttrMixin, HTMLTitleMixin
from .requests import GeoJSONType, JSONType, json_request

T = TypeVar("T", bound="Nominatim")


class Nominatim(GeoObject, HBoxMixin, HTMLTitleMixin, HTMLAttrMixin):
    """A class to parse Nominatim results.

    A Nominatim object is built based on JSON results of Nominatim requests.
    Nominatim requests are based on corresponding class methods:

    - Nominatim.search performs a search based on text;
    - Nominatim.reverse performs a search based on latlon coordinates;
    - Nominatim.lookup performs a search based on an OSM identifier.
    """

    shape = OrientedShape()

    endpoint = "https://nominatim.openstreetmap.org/"
    html_attr_list = [
        "osm_type",
        "osm_id",
        "address",
        "category",
        "type_",
        "importance",
    ]

    def __init__(self, json: JSONType) -> None:
        super().__init__()
        self.json = json
        self.shape = shape(self.json["geojson"])

    @property
    def __geo_interface__(self) -> GeoJSONType:
        return mapping(self.shape)

    @property
    def simple_json(self) -> JSONType:
        return {
            "place_id": self.json.get("place_id", None),
            "display_name": self.json.get("display_name", None),
            "lat": round(float(self.json.get("lat", "nan")), 5),
            "lon": round(float(self.json.get("lon", "nan")), 5),
        }

    def __repr__(self) -> str:
        return f"{type(self).__name__} {self.simple_json}"

    def __getattr__(self, name):
        if name.endswith("_"):  # in case the name is reserved
            name = name[:-1]
        value = self.json.get(name, None)
        if value is None:
            address = self.json.get("address", None)
            if address is not None:
                value = address.get(name, None)
        if value is None:
            raise AttributeError(name)
        return value

    @property
    def address(self) -> str:
        address = self.json.get("address", None)
        if address:
            return address
        return self.json["display_name"].split(", ")

    def _repr_html_(self) -> str:
        return (
            super()._repr_html_()
            + "<div style='float: left; margin: 10px;'>"
            + self._repr_svg_()
            + "</div>"
        )

    @classmethod
    def search(cls: Type[T], name: str, **kwargs) -> Optional[T]:
        """Performs a Nominatim search request.

        The request is based on the name passed in parameter.
        >>> Nominatim.search("Toulouse")
        Nominatim {'place_id': 256863032, 'display_name': 'Toulouse, ...', 'lat': 43.60446, 'lon': 1.44425}

        """
        params = dict(
            q=name,
            format="jsonv2",
            limit=1,
            dedupe=False,
            polygon_geojson=True,
            addressdetails=True,
        )
        json = json_request(
            cls.endpoint.rstrip("/") + "/" + "search",
            timeout=30,
            params=params,
            **kwargs,
        )
        if len(json) == 0:
            return None
        return cls(json[0])

    @classmethod
    def reverse(
        cls: Type[T], latitude: float, longitude: float, **kwargs
    ) -> Optional[T]:
        """Performs a Nominatim search request.

        The request is based on the latlon coordinates of the element.
        >>> Nominatim.reverse(43.608, 1.442)
        Nominatim {'place_id': 155591309, 'display_name': 'Basilique Saint-Sernin, ...', 'lat': 43.6..., 'lon': 1.44...}
        """

        params = dict(
            lat=latitude,
            lon=longitude,
            format="jsonv2",
            polygon_geojson=True,
        )
        json = json_request(
            cls.endpoint.rstrip("/") + "/" + "reverse",
            timeout=30,
            params=params,
            **kwargs,
        )
        if len(json) == 0:
            return None
        return cls(json)

    @classmethod
    def lookup(
        cls: Type[T], osm_ids: Union[str, List[str]], **kwargs
    ) -> Union[None, T, List[T]]:
        """Performs a Nominatim search request.

        The request is based on the OSM id of the element. The prefix
        determines the type of the OSM object (N for node, W for way, R for
        relation)
        >>> Nominatim.lookup("R367073")
        Nominatim {'place_id': 256948794, 'display_name': 'Capitole ...', 'lat': 43.60445, 'lon': 1.44449}
        """
        params = dict(
            osm_ids=osm_ids,
            format="jsonv2",
            polygon_geojson=True,
        )
        json = json_request(
            cls.endpoint.rstrip("/") + "/" + "lookup",
            timeout=30,
            params=params,
            **kwargs,
        )
        if len(json) == 0:
            return None
        elif len(json) == 1:
            return cls(json[0])
        else:
            return list(cls(json_ for json_ in json))



import itertools
import logging
from operator import itemgetter

from shapely.geometry import MultiPolygon as ShapelyMPL
from shapely.geometry import Polygon
from shapely.geometry.collection import GeometryCollection
from shapely.geometry.multilinestring import MultiLineString
from shapely.ops import linemerge, polygonize, unary_union

from .. import Overpass
from ..core import Relation

_log = logging.getLogger(__name__)


class MultiPolygon(Relation):
    """A class to parse multipolygon=* relations.

    Relations of type multipolygon are used to represent complex areas with
    holes inside or consisting of multiple disjoint areas.

    Reference: https://wiki.openstreetmap.org/wiki/Relation:multipolygon

    Tags:
      - type (multipolygon)
      - natural
      - landuse
      - building
      - man_made
      - amenity
      - leisure
      - highway (pedestrian)
      - waterway
      - ...

    Relation members:
      - outer 1+
      - inner 0+

    """

    def __init__(self, json):

        super().__init__(json)

        self.parent: Overpass = json["_parent"]
        parsed_keys = dict(
            (elt["ref"], elt)
            for elt in self.parent.all_members[json["id_"]]
            if elt is not None
        )

        parts = dict(
            (
                role,
                unary_union(
                    list(
                        elt["geometry"]
                        for elt in it
                        # labels do not have a geometry for instance...
                        if elt.get("geometry", None)
                    )
                ),
            )
            for role, it in itertools.groupby(
                parsed_keys.values(), key=itemgetter("role")
            )
        )

        if "outer" not in parts:  # LEBL: Terminal 2
            msg = f"Invalid geometry with id {json['id_']}: no outer element"
            _log.warning(msg)
            self.json["geometry"] = self.shape = GeometryCollection()
            return

        if isinstance(parts["outer"], ShapelyMPL):
            parts["outer"] = unary_union(
                list(Polygon(part) for part in parts["outer"].geoms)
            )

        if isinstance(parts["outer"], MultiLineString):  # LFPO, LFPG
            parts["outer"] = linemerge(parts["outer"])
            if isinstance(parts["outer"], MultiLineString):  # again, give up...
                self.json["geometry"] = ShapelyMPL(polygonize(parts["outer"]))
                self.shape = self.json["geometry"]
                return

        if parts.get("inner", None):
            if isinstance(parts["inner"], MultiLineString):
                parts["inner"] = parts["inner"].geoms
            else:  # EDDF
                # LineString has not attribute geoms...
                parts["inner"] = [parts["inner"]]

        msg = (
            f"Error parsing multigeometry with id {json['id_']}. "
            "Falling back to Polygon without holes."
        )
        try:
            self.shape = Polygon(
                shell=getattr(parts["outer"], "geoms", parts["outer"]),
                holes=parts.get("inner", None),
            )
            self.json["geometry"] = self.shape

        except Exception as msg:
            _log.warning(
                f"Error parsing multigeometry with id {json['id_']}: {msg} "
            )
            self.json["geometry"] = self.shape = GeometryCollection()


class Building(MultiPolygon):
    """Should be deprecated."""



import logging
from concurrent.futures import Future, ProcessPoolExecutor, as_completed
from functools import lru_cache
from operator import itemgetter
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    TypedDict,
)

import geopandas as gpd
from tqdm.autonotebook import tqdm

import pandas as pd
from pyproj import Proj
from shapely.geometry.base import BaseGeometry
from shapely.ops import unary_union

from ...crs import PlateCarree  # type: ignore
from ...dataviz import matplotlib_style
from ...utils.cache import CacheResults, cached_property
from ...utils.descriptors import Descriptor
from ...utils.geometry import reorient
from ..requests import JSONType, json_request
from .core import NodeWayRelation, to_geometry
from .query import Query

_log = logging.getLogger(__name__)


class Member(TypedDict):
    ref: int
    role: str
    geometry: Optional[BaseGeometry]


def hashing_id(*args: Dict[str, int], **kwargs: Dict[str, int]) -> int:
    elt = args[1] if len(args) > 1 else kwargs["elt"]
    return elt["id"]


cache_by_id = CacheResults(
    cache_dir=".",
    hashing=hashing_id,
    # no storing in files
    writer=lambda _a, _b: None,
    reader=lambda _: dict(unused=True),
)


class OverpassDataDescriptor(Descriptor[gpd.GeoDataFrame]):
    """Builds the GeoDataFrame on demand.
    Validates it has required fields when replaced.
    """

    def __get__(self, obj, cls=None) -> gpd.GeoDataFrame:
        data = getattr(obj, self.private_name)
        if data is None:
            data = obj.parse()
            for col in ["_parent", "members", "nodes"]:
                if col in data.columns:
                    data = data.drop(columns=col)
            setattr(obj, self.private_name, data)
        if "latitude" in data.columns:
            x = data.query("latitude != latitude")
        else:
            x = data.assign(latitude=None, longitude=None)

        if x.shape[0] > 0 and "geometry" in x.columns:
            x = x.query("geometry == geometry and not geometry.is_empty")
            if x.shape[0] > 0:
                data.loc[x.index, ["latitude", "longitude"]] = x.assign(
                    longitude=lambda df: df.geometry.centroid.x,
                    latitude=lambda df: df.geometry.centroid.y,
                )[["latitude", "longitude"]]

        if "geometry" in data.columns:
            data = data.assign(
                # TODO This should not be necessary, yet...
                geometry=lambda df: df.geometry.apply(reorient)
            )

        return data.dropna(axis=1, how="all")

    def __set__(self, obj, data: gpd.GeoDataFrame) -> None:
        feat = ["id_", "type_", "geometry"]
        msg = f"Did you remove any of {feat} from {data.columns}?"
        if any(f not in data.columns for f in feat):
            raise TypeError(msg)

        setattr(obj, self.private_name, data)


class Overpass:
    endpoint = "http://www.overpass-api.de/api/interpreter"
    data = OverpassDataDescriptor()

    def __init__(self, json: JSONType, data: Optional[gpd.GeoDataFrame] = None):
        super().__init__()
        self.json = json
        self._bounds: Optional[Tuple[float, float, float, float]] = None
        if data is None:
            self._data = None
        else:
            self.data = data

    def _repr_html_(self):
        return self.data._repr_html_()

    def __getstate__(self):
        """
        In order for the multiprocessing step to be efficient, we choose here
        to empty the JSON and avoid numerous useless pickling.

        Currently used only in .simplify()
        """
        # Compute it once, because we need it for the simplification process
        if getattr(self, "simplify_flag", None):
            _ = self.bounds, self.data
        # Copy the object's state from self.__dict__ which contains
        # all our instance attributes. Always use the dict.copy()
        # method to avoid modifying the original state.
        state = self.__dict__.copy()
        # Usually, we do not need the json of the parent for the simplification
        if getattr(self, "simplify_flag", None):
            state["json"] = dict(
                (key, value)
                for key, value in self.json.items()
                if key != "elements"
            )
        return state

    def __setstate__(self, state):
        # Restore instance attributes
        self.__dict__.update(state)

    @property
    def bounds(self) -> Tuple[float, float, float, float]:
        if self._bounds is not None:
            return self._bounds
        if "geometry" in self.data.columns:
            x = (
                self.data.query("geometry == geometry")
                .assign(empty=lambda df: df.geometry.is_empty)
                .query("not empty")
            )
            if x.shape[0] > 0:
                self._bounds = unary_union(x.geometry).bounds
                return self._bounds
        self._bounds = tuple(  # type: ignore
            eval(key[:3])(
                (x["bounds"] for x in self.json["elements"] if "bounds" in x),
                key=itemgetter(key),
            )[key]
            for key in ["minlon", "minlat", "maxlon", "maxlat"]
        )
        return self._bounds  # type: ignore

    @property
    def extent(self) -> Tuple[float, float, float, float]:
        west, south, east, north = self.bounds
        return west, east, south, north

    @classmethod
    def request(
        cls, query: Optional[str] = None, *args, **kwargs
    ) -> "Overpass":
        if query is None:
            query = cls.build_query(*args, **kwargs)
        res = json_request(url=Overpass.endpoint, data=query)
        return Overpass(res)

    @staticmethod
    def build_query(*, out: str = "json", timeout: int = 180, **kwargs) -> str:
        query = Query(out=out, timeout=timeout, **kwargs)
        return query.generate()

    def assign(self, *args, **kwargs) -> "Overpass":
        return Overpass(self.json, self.data.assign(*args, **kwargs))

    def head(self, *args, **kwargs) -> "Overpass":
        return Overpass(self.json, self.data.head(*args, **kwargs))

    def tail(self, *args, **kwargs) -> "Overpass":
        return Overpass(self.json, self.data.tail(*args, **kwargs))

    def query(self, *args, **kwargs) -> "Overpass":
        return Overpass(self.json, self.data.query(*args, **kwargs))

    def drop(self, *args, **kwargs) -> "Overpass":
        return Overpass(self.json, self.data.drop(*args, **kwargs))

    def merge(self, *args, **kwargs) -> "Overpass":
        return Overpass(self.json, self.data.merge(*args, **kwargs))

    def dropna(self, *args, **kwargs) -> "Overpass":
        return Overpass(self.json, self.data.dropna(*args, **kwargs))

    def sort_values(self, *args, **kwargs) -> "Overpass":
        return Overpass(self.json, self.data.sort_values(*args, **kwargs))

    def area(self) -> "Overpass":
        bounds = self.bounds
        proj = Proj(
            proj="aea",  # equivalent projection
            lat_1=bounds[1],
            lat_2=bounds[3],
            lat_0=(bounds[1] + bounds[3]) / 2,
            lon_0=(bounds[0] + bounds[2]) / 2,
        )
        return self.assign(
            area=self.data.geometry.set_crs(epsg=4326).to_crs(crs=proj.crs).area
        )

    def length(self) -> "Overpass":
        bounds = self.bounds
        proj = Proj(
            proj="aea",  # equivalent projection
            lat_1=bounds[1],
            lat_2=bounds[3],
            lat_0=(bounds[1] + bounds[3]) / 2,
            lon_0=(bounds[0] + bounds[2]) / 2,
        )
        return self.assign(
            length=self.data.geometry.set_crs(epsg=4326)
            .to_crs(crs=proj.crs)
            .length
        )

    def coloring(self) -> "Overpass":
        import networkx as nx

        self.graph = nx.Graph()
        for elt in self:
            for neighbour in elt.neighbours:
                self.graph.add_edge(elt.json["id_"], neighbour)
        colors = nx.algorithms.coloring.greedy_color(self.graph)
        merge = self.merge(
            pd.Series(colors, name="coloring"), left_on="id_", right_index=True
        )
        merge.graph = self.graph
        return merge

    def simplify(
        self,
        resolution: Optional[float] = None,
        max_workers: Optional[int] = 4,
        **kwargs,
    ) -> "Overpass":

        if resolution is None:
            return self

        new_overpass = Overpass(self.json)

        if max_workers is not None and max_workers > 1:
            self.simplify_flag = True
            with ProcessPoolExecutor(max_workers=max_workers) as executor:
                futures: Dict[Future, int] = dict()
                for i, nwr in enumerate(self):
                    futures[
                        executor.submit(nwr.simplify, resolution, **kwargs)
                    ] = i
                result: List[Tuple[int, BaseGeometry]] = list()
                for future in tqdm(as_completed(futures), total=len(futures)):
                    result.append((futures[future], future.result().shape))
            self.simplify_flag = False
        else:
            result = list(
                (i, nwr.simplify(resolution, **kwargs))
                for i, nwr in tqdm(enumerate(self), total=self.data.shape[0])
            )

        new_overpass.data = self.data.assign(
            geometry=list(elt for _, elt in sorted(result))
        )
        new_overpass.json = self.json

        return new_overpass

    def __iter__(self) -> Iterator[NodeWayRelation]:
        for _, line in self.data.iterrows():
            yield NodeWayRelation({"_parent": self, **dict(line)})

    @lru_cache()
    def __getitem__(self, item) -> NodeWayRelation:
        elt = self.data.query("id_ == @item")
        if elt.shape[0] == 0:
            raise AttributeError(f"No {item} id_ in the current data")
        return NodeWayRelation({"_parent": self, **dict(elt.iloc[0])})

    @property
    def __geo_interface__(self):
        return self.data.__geo_interface__

    @cache_by_id
    def make_node(self, elt: Dict[str, Any]) -> Dict[str, Any]:
        return dict(
            _parent=self,
            id_=elt["id"],
            type_=elt["type"],
            latitude=elt["lat"],
            longitude=elt["lon"],
            geometry=to_geometry(elt) if elt.get("geometry", None) else None,
            **elt["tags"],
        )

    @cache_by_id
    def make_way(self, elt: Dict[str, Any]) -> Dict[str, Any]:
        return dict(
            _parent=self,
            id_=elt["id"],
            type_=elt["type"],
            nodes=elt["nodes"],
            geometry=to_geometry(elt) if elt.get("geometry", None) else None,
            **elt["tags"],
        )

    @cache_by_id
    def make_relation(self, elt: Dict[str, Any]) -> Dict[str, Any]:

        return NodeWayRelation(
            dict(
                _parent=self,
                id_=elt["id"],
                type_=elt["type"],
                members=elt["members"],
                **elt["tags"],
            )
        ).json

    @cached_property
    def all_members(self) -> Dict[int, List[Member]]:
        return dict(
            (
                elt["id"],
                list(
                    {
                        "ref": entry["ref"],
                        "role": entry["role"],
                        "geometry": to_geometry(entry)
                        if entry.get("geometry", None)
                        else None,
                    }
                    for entry in elt.get("members", [])
                    if entry["type"] != "relation"
                    # TODO subrelations not supported
                ),
            )
            for elt in self.json["elements"]
        )

    @lru_cache()
    def parse(self) -> gpd.GeoDataFrame:
        make_dict: Dict[
            str, Callable[[Overpass, Dict[str, Any]], Dict[str, Any]]
        ] = {
            "node": Overpass.make_node,
            "way": Overpass.make_way,
            "relation": Overpass.make_relation,
        }
        return gpd.GeoDataFrame.from_records(
            list(
                make_dict[elt["type"]](self, elt)
                for elt in self.json["elements"]
                if elt.get("tags", None)
            )
        )

    def plot(self, ax, by: Optional[str] = None, **kwargs):
        if by is None:
            return self.data.plot(ax=ax, transform=PlateCarree(), **kwargs)
        for key, elt in self.data.groupby(by):
            if (
                by not in matplotlib_style
                and key not in matplotlib_style
                and by not in kwargs
                and key not in kwargs
            ):
                _log.warning(f"{by}={key} not in stylesheet, hence ignored")
                continue
            current_style = {
                **matplotlib_style.get(by, {}),
                **matplotlib_style.get(key, {}),
                **kwargs.get(by, {}),
                **kwargs.get(key, {}),
            }
            if "key" in current_style:
                if current_style["key"] not in elt.columns:
                    continue
                for cat, piece in elt.groupby(current_style["key"]):
                    if cat in matplotlib_style[key]:
                        piece.plot(
                            ax=ax,
                            transform=PlateCarree(),
                            **current_style.get(cat, {}),
                        )
                    else:
                        msg = f"{key}={cat} not in stylesheet, hence ignored"
                        _log.warning(msg)
            else:
                elt.plot(ax=ax, transform=PlateCarree(), **current_style)



import logging
import re
from abc import ABC, abstractmethod
from numbers import Real
from typing import Any, Dict, List, Mapping, Optional, Union

from .. import Nominatim

QueryType = Union[bool, int, str, List, Mapping[str, Any]]

_log = logging.getLogger(__name__)


class Generator(ABC):
    def __set_name__(self, owner, name):
        self.public_name = name
        self.private_name = "_" + name

    def __get__(self, obj, objtype=None) -> str:
        elt = getattr(obj, self.private_name, None)
        if elt is None:
            return ""
        if not obj.geometry:
            _log.info(f"{self.public_name} called with geom={obj.geometry}")
        return self.generate(elt, obj, geom=obj.geometry)

    def __set__(self, obj, value):
        if value is not None:
            value = self.transform(value)
            self.validate(value)
        setattr(obj, self.private_name, value)

    def transform(self, value) -> Any:
        return value

    @abstractmethod
    def validate(self, value):
        pass

    @abstractmethod
    def generate(self, elt, obj, geom: bool = True) -> str:
        pass


def expand(key: str, value: Union[bool, str, Dict[str, str]], op=dict(op="=")):
    if not re.match(r"^[\w\d_]+$", key):
        key = f'"{key}"'
    if isinstance(value, bool):
        return f"[{key}]"
    if isinstance(value, Real):
        return f"[{key}{op['op']}{value}]"
    if isinstance(value, str):
        if not re.match(r"^[\w\d_]+$", value):
            value = f'"{value}"'
        return f"[{key}{op['op']}{value}]"
    if isinstance(value, str):
        if not re.match(r"^[\w\d_]+$", value):
            value = f'"{value}"'
        return f"[{key}{op['op']}{value}]"
    if isinstance(value, list):
        return f'[{key}{op["op"]}"{"|".join(value)}"]'
    if "op" in value:
        return expand(key, value["value"], op=dict(op=value["op"]))
    if "regex" in value:
        value = value["regex"]
        return f'[{key}{op.get("fuzzy", "~")}"{value}"]'
    raise ValueError(f"Unknown elements in dictionary {value}")


class Area(Generator):
    def transform(self, value):
        if isinstance(value, str):
            value = Nominatim.search(value)
        return value

    def validate(self, value):
        if isinstance(value, Nominatim):
            return
        if not (isinstance(value, dict)):
            msg = "The area parameter must be a dictionary or a Nominatim"
            raise TypeError(msg)
        for elt in value.values():
            if not any(isinstance(elt, t) for t in [Real, str, dict]):
                raise TypeError(
                    "All values in the area dictionary must be either "
                    "numeric, str or a dict[str, str]"
                )

    def generate(self, elt, obj, geom: bool = True) -> str:
        res = ";"
        if isinstance(elt, dict):
            if "as_" in elt:
                res = f"->.{elt['as_']};"
                del elt["as_"]
            return (
                "area"
                + "".join(expand(key, value) for key, value in elt.items())
                + res
            )
        elif isinstance(elt, Nominatim):
            type_ = elt.json["osm_type"]
            if type_ == "relation":
                type_ = "rel"
            return f"{type_}(id:{elt.json['osm_id']});map_to_area" + res
        msg = "The area parameter must be a dictionary or a Nominatim"
        raise TypeError(msg)


class Bounds(Generator):
    def transform(self, value):
        if hasattr(value, "shape"):
            value = value.shape.bounds
        return value

    def validate(self, value):
        try:
            west, south, east, north = value
            assert all(isinstance(x, Real) for x in value)
        except Exception as exc:
            _log.warning(exc)
            raise TypeError(
                "The bounds attribute must be a shape object or "
                "a tuple of four float (west, south, east, north)"
            )

    def generate(self, value, obj, geom: bool = True) -> str:
        west, south, east, north = value
        return f"[bbox:{south},{west},{north},{east}]"


class NodeWayRel(Generator):
    def transform(self, value):
        if isinstance(value, dict):
            if not any(x in value for x in ["node", "way", "rel", "nwr"]):
                value = dict(nwr=value)
            value = [value]
        return value

    def validate(self, value):
        pass

    def generate(self, value, obj, geom: bool = True) -> str:
        return "".join(self.generate_single(elt, obj, geom) for elt in value)

    def generate_single(self, elt, obj, geom: bool) -> str:
        geom_str = " geom" if geom else ""
        for res, elt in elt.items():
            break
        # Particular situation of pivot relations
        if list(elt.keys()) == ["area"]:
            if elt["area"] is True:
                res += f"(pivot);out{geom_str};"
            else:
                res += f"(pivot.{elt['area']});out{geom_str};"
            return res
        if getattr(obj, "_area", None) is not None:
            area_name = elt.get("area", None)
            if area_name:
                del elt["area"]
                res += f"(area.{area_name})"
            else:
                res += "(area)"
        if "around" in elt:
            res += f"(around:{elt['around']})"
            del elt["around"]
        res += "".join(expand(key, value) for key, value in elt.items())
        res += f";out{geom_str};"
        return res


def make_querytype(
    values: Optional[QueryType] = None, key="nwr"
) -> List[QueryType]:
    if values is None:
        return []
    if isinstance(values, list):
        return list(
            (
                {key: elt}
                if not any(
                    x in elt.keys() for x in ["node", "way", "rel", "nwr"]
                )
                else elt
            )
            for elt in values
        )
    else:
        return [{key: values}]


class Query:
    area = Area()
    bounds = Bounds()
    nwr = NodeWayRel()

    def __init__(
        self,
        *,
        out: str = "json",
        timeout: int = 180,
        geometry: bool = True,
        **kwargs: QueryType,
    ) -> None:
        self.out = out
        self.timeout = timeout
        self.geometry = geometry
        self.area = kwargs.get("area", None)
        self.bounds = kwargs.get("bounds", None)
        self.date = kwargs.get("date", None)

        if not geometry:
            _log.warning("geometry=False functionality still experimental")

        node = make_querytype(kwargs.get("node", None), "node")
        way = make_querytype(kwargs.get("way", None), "way")
        rel = make_querytype(kwargs.get("rel", None), "rel")
        nwr = make_querytype(kwargs.get("nwr", None), "nwr")

        kwargs = dict(
            (key, value)
            for key, value in kwargs.items()
            if key
            not in ["area", "bounds", "date", "node", "way", "rel", "nwr"]
        )
        if len(kwargs):
            nwr.append(dict(nwr=kwargs))

        self.nwr = nwr + node + way + rel

    def generate(self) -> str:
        res = (
            f"[out:{self.out}][timeout:{self.timeout}]"
            + (f"[date:'{self.date}']" if self.date is not None else "")
            + f"{self.bounds};"
        )
        res += f"{self.area}"
        res += f"{self.nwr}"
        return res



import itertools
from collections import UserDict
from operator import itemgetter
from typing import Dict, Iterator, List, Set, Tuple, TypeVar, Union, cast

from pyproj import Proj, Transformer
from shapely.geometry import MultiLineString, MultiPolygon, Polygon
from shapely.geometry.base import BaseGeometry
from shapely.geometry.collection import GeometryCollection
from shapely.ops import polygonize, transform, unary_union

from ....utils.cache import cached_property
from ....utils.geometry import reorient
from .. import Overpass
from ..core import Relation


class RelationsDict(UserDict):
    def __missing__(self, key):
        value = self[key] = list()
        return value

    def include(self, chunk: BaseGeometry, role: str):
        if isinstance(chunk, MultiLineString):
            for c in chunk.geoms:
                self[role].append(c)
        else:
            self[role].append(chunk)


T = TypeVar("T", bound="Boundary")


class Boundary(Relation):
    """A class to parse boundary=* relations.

    boundary=* relations are used for grouping boundaries and marking
    enclaves/exclaves.

    Reference: https://wiki.openstreetmap.org/wiki/Relation:boundary

    Tags:
      - type (boundary)
      - boundary (administrative)
      - land_area (administrative)
      - name
      - admin_level

    Relation members:
      - outer 1+
      - inner 0+
      - admin_center 0-1
      - label 0-1
      - subarea 0+
    """

    def __init__(self, json):

        super().__init__(json)

        self.known_chunks = RelationsDict()
        self.parent: Overpass = json["_parent"]

        self.parsed_keys = cast(
            Dict[int, Dict[str, Union[int, str]]],
            dict(
                (elt["ref"], elt)
                for elt in self.parent.all_members[self.json["id_"]]
            ),
        )

        self._build_geometry_parts()
        if len(self.known_chunks) > 0:
            self._make_geometry(self.known_chunks)

    def simplify(
        self: T,
        resolution: float,
        bounds: Union[None, Tuple[float, float, float, float]] = None,
    ) -> T:
        if bounds is None:
            bounds = self.parent.bounds
        b0, b1, b2, b3 = bounds

        def simplify_shape(shape_: BaseGeometry) -> BaseGeometry:
            proj = Proj(
                proj="aea",  # equivalent projection
                lat_1=b1,
                lat_2=b3,
                lat_0=(b1 + b3) / 2,
                lon_0=(b0 + b2) / 2,
            )

            forward = Transformer.from_proj(
                Proj("EPSG:4326"), proj, always_xy=True
            )
            backward = Transformer.from_proj(
                proj, Proj("EPSG:4326"), always_xy=True
            )
            return transform(
                backward.transform,
                transform(forward.transform, shape_).simplify(resolution),
            )

        rel_dict = RelationsDict()
        for role, shapes in self.known_chunks.items():
            for elt in shapes:
                rel_dict.include(simplify_shape(elt), role)

        self._make_geometry(rel_dict)
        return self

    def _make_geometry(self, parts: RelationsDict):

        outer: List[Polygon] = list(polygonize(parts["outer"]))
        inner: List[Polygon] = list(polygonize(parts["inner"]))

        if len(outer) == 0:
            return GeometryCollection()

        if len(outer) > 1:
            list_ = [
                Polygon(o, holes=list(i for i in inner if o.contains(i)))
                for o in outer
            ]
            shape = MultiPolygon(list_)

        else:
            shape = Polygon(outer[0], inner)

        self.shape = reorient(shape)
        self.json["geometry"] = self.shape

    def intersections(self) -> Iterator[Tuple[int, Set[int]]]:
        all_sets: Dict[int, Set[int]] = dict(
            (key, set(e["ref"] for e in list_))
            for key, list_ in self.parent.all_members.items()
        )
        key_i = self.json["id_"]
        set_i = all_sets[key_i]
        for key_j, set_j in all_sets.items():
            if key_i == key_j:
                continue
            intersection = set_i.intersection(set_j)
            if len(intersection) > 0:
                yield key_j, intersection

    @cached_property
    def neighbours(self) -> Set[int]:
        return set(key for key, _value in self.intersections())

    def include(self, elements: Set[int]) -> None:
        for role, it in itertools.groupby(
            (self.parsed_keys[key] for key in elements), key=itemgetter("role")
        ):
            elts = list(
                elt["geometry"] for elt in it if elt["geometry"] is not None
            )
            if len(elts) > 0:
                chunk = unary_union(elts)
                self.known_chunks.include(chunk, role)

    def _build_geometry_parts(self) -> BaseGeometry:

        all_: Set[int] = set(self.parsed_keys.keys())

        for _id, intersection in self.intersections():
            self.include(intersection)
            all_ -= intersection

        self.include(all_)


# Forcing the inheritance for proper registration
class Land_Area(Boundary):
    pass



import hashlib
import json
import logging
import time
from io import BytesIO
from pathlib import Path
from typing import Any, Optional

import requests
from appdirs import user_cache_dir
from tqdm.autonotebook import tqdm

from ..utils.cache import CacheResults

JSONType = Any
GeoJSONType = Any
session = requests.Session()

_log = logging.getLogger(__name__)


def _hash_request(*args, **kwargs) -> str:
    if "timeout" in kwargs:
        del kwargs["timeout"]

    if "data" in kwargs:  # overpass requests
        query = kwargs["data"].replace("\n", "").replace(" ", "")
        hashcode = hashlib.md5(query.encode("utf-8")).hexdigest()
    else:  # usual ones
        prepared_url = requests.Request("POST", *args, **kwargs).prepare().url
        assert prepared_url is not None
        hashcode = hashlib.md5(prepared_url.encode("utf-8")).hexdigest()
    return hashcode + ".json"


def _write_json(json_: JSONType, cache_file) -> None:  # coverage: ignore
    _log.info(f"Writing cache file {cache_file}")
    cache_file.write_text(json.dumps(json_, indent=2))


def _read_json(cache_file: Path) -> Optional[JSONType]:
    _log.info(f"Reading cache file {cache_file}")
    json_ = json.loads(cache_file.read_text())

    if isinstance(json_, list):
        return json_

    elements = json_.get("elements", None)
    if elements is None:
        return json_

    # Detect past failed cached Overpass queries
    _log.debug(f"Number of elements in cached results: {len(elements)}")
    if len(elements) == 0:
        return None

    return json_


@CacheResults(
    cache_dir=Path(user_cache_dir("cartes")) / "osm",
    hashing=_hash_request,
    reader=_read_json,
    writer=_write_json,
)
def json_request(url: str, timeout: int = 180, **kwargs) -> JSONType:
    """
    Send a request to the Overpass API and return the JSON response.
    """
    _log.info(f"Sending POST request to {url} with {kwargs}")

    new_kwargs = kwargs.copy()
    if "data" in new_kwargs:
        new_kwargs["data"] = new_kwargs["data"].encode("utf-8")

    response = session.post(url=url, timeout=timeout, **new_kwargs)

    if response.status_code == 504:  # gateway timeout
        msg = f"Got status code {response.status_code}. Trying again soon..."
        _log.warning(msg)
        time.sleep(5)
        return json_request(url, timeout=timeout, **kwargs)

    if response.status_code == 429:  # too many requests
        status = session.get("https://overpass-api.de/api/status")
        status.raise_for_status()
        print(status.content.decode())

    response.raise_for_status()

    b = BytesIO()

    if kwargs.get("stream", None):
        total_size = int(response.headers.get("content-length", 0))
        _log.info(f"{response.headers}")
        block_size = 1024 * 1024
        pbar = tqdm(total=total_size, unit="B", unit_scale=True)
        for data in response.iter_content(block_size):
            pbar.update(len(data))
            b.write(data)
        b.seek(0)
        pbar.close()
        if total_size != 0 and pbar.n != total_size:
            msg = f"Download not complete: {pbar.n}/{total_size}"
            raise RuntimeError(msg)

    try:
        if kwargs.get("stream", None):
            response_json = json.loads(b.read().decode())
        else:
            response_json = response.json()
    except Exception:
        msg = f"""Server returned no JSON data.
        {response} {response.reason}
        {response.text}"""
        _log.warning(msg)
        raise

    return response_json



from typing import Any

from cartopy import crs


class AlbersEqualArea(crs.AlbersEqualArea):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "conicEqualArea"
        if index == "rotate":
            lat = self.proj4_params["lat_0"]
            lon = self.proj4_params["lon_0"]
            return (-lon, -lat, 0)
        if index == "parallels":
            lat_1 = self.proj4_params["lat_1"]
            lat_2 = self.proj4_params["lat_2"]
            return [lat_1, lat_2]
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate", "parallels"]


class AzimuthalEquidistant(crs.AzimuthalEquidistant):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "azimuthalEquidistant"
        if index == "rotate":
            lat = self.proj4_params["lat_0"]
            lon = self.proj4_params["lon_0"]
            return (-lon, -lat, 0)
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate"]


class EqualEarth(crs.EqualEarth):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "equalEarth"
        if index == "rotate":
            lon = self.proj4_params["lon_0"]
            return (-lon, 0, 0)
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate"]


class EquidistantConic(crs.EquidistantConic):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "conicEquidistant"
        if index == "rotate":
            lat = self.proj4_params["lat_0"]
            lon = self.proj4_params["lon_0"]
            return (-lon, -lat, 0)
        if index == "parallels":
            lat_1 = self.proj4_params["lat_1"]
            lat_2 = self.proj4_params["lat_2"]
            return [lat_1, lat_2]
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate", "parallels"]


class Gnomonic(crs.Gnomonic):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "gnomonic"
        if index == "rotate":
            lat = self.proj4_params["lat_0"]
            lon = self.proj4_params["lon_0"]
            return (-lon, -lat, 0)
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate"]


class Mercator(crs.Mercator):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "mercator"
        if index == "rotate":
            lon = self.proj4_params["lon_0"]
            return (-lon, 0, 0)
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate"]


class Mollweide(crs.Mollweide):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "mollweide"
        if index == "rotate":
            lon = self.proj4_params["lon_0"]
            return (-lon, 0, 0)
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate"]


class LambertAzimuthalEqualArea(crs.LambertAzimuthalEqualArea):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "azimuthalEqualArea"
        if index == "rotate":
            lat = self.proj4_params["lat_0"]
            lon = self.proj4_params["lon_0"]
            return (-lon, -lat, 0)
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate"]


class LambertConformal(crs.LambertConformal):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "conicConformal"
        if index == "rotate":
            lat = self.proj4_params["lat_0"]
            lon = self.proj4_params["lon_0"]
            return (-lon, -lat, 0)
        if index == "parallels":
            lat_1 = self.proj4_params["lat_1"]
            lat_2 = self.proj4_params["lat_2"]
            return [lat_1, lat_2]
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate", "parallels"]


class Orthographic(crs.Orthographic):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "orthographic"
        if index == "rotate":
            lat = self.proj4_params["lat_0"]
            lon = self.proj4_params["lon_0"]
            return (-lon, -lat, 0)
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate"]


class PlateCarree(crs.PlateCarree):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "equirectangular"
        raise KeyError(index)

    def keys(self):
        return ["type"]


class Stereographic(crs.Stereographic):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "stereographic"
        if index == "rotate":
            lat = self.proj4_params["lat_0"]
            lon = self.proj4_params["lon_0"]
            return (-lon, -lat, 0)
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate"]


class TransverseMercator(crs.TransverseMercator):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "transverseMercator"
        if index == "rotate":
            lat = self.proj4_params["lat_0"]
            lon = self.proj4_params["lon_0"]
            return (-lon, -lat, 0)
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate"]


class UTM(crs.UTM):
    def __getitem__(self, index: str) -> Any:
        if index == "type":
            return "transverseMercator"
        if index == "rotate":
            zone = self.proj4_params["zone"]
            return (-(zone * 6 - 180 - 3), 0, 0)
        raise KeyError(index)

    def keys(self):
        return ["type", "rotate"]


class EuroPP(UTM, crs.EuroPP):
    pass


class OSGB(TransverseMercator, crs.OSGB):
    pass


class OSNI(TransverseMercator, crs.OSNI):
    pass



from __future__ import annotations

import asyncio
import json
import re
import time
from io import StringIO
from operator import attrgetter
from pathlib import Path
from typing import Any

import aiohttp
import altair as alt
import geopandas as gpd
import requests
from appdirs import user_cache_dir

import pandas as pd

from ..osm.requests import session
from ..utils.cache import (
    CacheResults,
    cached_property,
    read_json,
    read_json_df,
    write_df_json,
    write_json,
)
from ..utils.mixins import HTMLMixin


class GithubAPI:
    id_: str
    base_url = "https://api.github.com/repos/{username}/{repository}/contents/"

    def __init__(
        self,
        username: str = "deldersveld",
        repository: str = "topojson",
    ) -> None:
        self.base_url = GithubAPI.base_url.format(
            username=username, repository=repository
        )
        self.id_ = f"{username}_{repository}"

    async def async_get_recursive(
        self,
        session: aiohttp.ClientSession,
        path: str = "",
    ) -> pd.DataFrame:

        async with session.get(self.base_url + path) as resp:
            json = await resp.json()
            if "message" in json:
                raise RuntimeError(
                    f"{json['message']}\n{json['documentation_url']}"
                )

        df = pd.json_normalize(json)[["path", "download_url", "type", "size"]]

        futures = list(
            self.async_get_recursive(session, elt)
            for elt in df.query('type == "dir"').path
        )
        df_list = [df] + list(
            result
            for result in await asyncio.gather(*futures, return_exceptions=True)
        )

        return pd.concat(df_list)

    async def async_get_features(self) -> pd.DataFrame:
        async with aiohttp.ClientSession(trust_env=True) as s:
            return await self.async_get_recursive(s)

    def get_features(self) -> pd.DataFrame:
        loop = asyncio.get_event_loop()

        if loop.is_running():  # proxy for Jupyter notebook
            msg = "For details, run the same command inside a Python console."
            future = loop.create_task(self.async_get_features())
            for _ in range(5):
                if future.done():
                    result = future.result()
                    break
                time.sleep(1)
            else:
                future.cancel()
                raise TimeoutError(msg)
        else:
            result = asyncio.run(self.async_get_features())

        return result


class NpmAPI:
    api_url = "https://data.jsdelivr.com/v1/package/npm"
    base_url = "https://cdn.jsdelivr.net/npm"
    id_: str

    def __init__(self, name="world-atlas@2.0.2"):
        self.name = name
        if "@" not in name:
            c = requests.get(self.api_url + self.name)
            c.raise_for_status()
            self.latest = c.json()["tags"]["latest"]
        else:
            self.name, self.latest = name.split("@")
        self.id_ = f"{self.name}@{self.latest}"

    def get_features(self) -> pd.DataFrame:

        c = requests.get(f"{self.api_url}/{self.id_}/flat")
        c.raise_for_status()
        prefix = f"{self.base_url}/{self.id_}/"

        return (
            pd.json_normalize(c.json()["files"])
            .assign(name=lambda df: df.name.str[1:])  # remove initial /
            .assign(download_url=lambda df: prefix + df.name)
            .rename(columns=dict(name="path"))
            .drop(columns="time")
        )


@CacheResults(
    cache_dir=Path(user_cache_dir("cartes")) / "atlas",
    hashing=attrgetter("path"),
    reader=read_json,
    writer=write_json,
)
def get_json(elt: "TopoCatalogue") -> Any:
    c = session.get(elt.url)
    c.raise_for_status()
    return c.json()


class TopoCatalogue(HTMLMixin):
    def __init__(
        self,
        df: pd.DataFrame | None = None,
        api: GithubAPI | NpmAPI | None = None,
        username: str = "",
        repository: str = "",
    ) -> None:
        self.api = api
        if df is None:
            if self.api is None:
                self.api = GithubAPI(username=username, repository=repository)
            cache_dir = Path(user_cache_dir("cartes")) / "atlas" / self.api.id_
            get_features = CacheResults(
                cache_dir=cache_dir,
                hashing=lambda: "catalogue.json",
                reader=read_json_df,
                writer=write_df_json,
            )(self.api.get_features)
            df = get_features()

        self.df = df.query('path.str.endswith("json")')

    def __getattr__(self, name: str) -> "TopoCatalogue":
        if re.match("(_[0-9])?[A-Za-z0-9]+_?", name):
            if name.startswith("_"):  # escape numbers
                name = name[1:]
            res = self.q(name)
            if res is not None:
                return res
        raise AttributeError(name)

    def q(self, name: str) -> "TopoCatalogue" | None:
        if name.endswith("_"):
            query = f'not path.str.contains("{name[:-1]}")'
        else:
            query = f'path.str.contains("{name}")'
        new_data = self.df.query(query)
        if new_data.shape[0] > 0:
            return TopoCatalogue(new_data, api=self.api)
        return None

    def __repr__(self) -> str:
        return repr(self.df)

    def _repr_html_(self) -> str:
        return self.df._repr_html_()

    @property
    def url(self):
        return self.df.iloc[0].download_url

    @property
    def path(self):
        assert self.api is not None
        return Path(self.api.id_) / self.df.iloc[0].path

    @cached_property
    def features(self) -> list[str]:
        return list(get_json(self)["objects"].keys())

    @property
    def topo_feature(self):
        feature = self.features[0]
        return alt.topo_feature(self.url, feature=feature)

    @property
    def data(self):
        feature = self.features[0]
        return gpd.read_file(
            StringIO(json.dumps(get_json(self))), layout=feature
        )



import logging

import click

from cartes import crs
from cartes.utils.features import countries


@click.command()
@click.argument("projection", type=str, default="Mercator")
@click.option(
    "--scale",
    type=str,
    default="50m",
    show_default=True,
    help="Resolution for country outlines",
)
@click.option("-v", "--verbose", count=True, help="Verbosity level")
def main(projection: str, scale: str, verbose: int):

    logger = logging.getLogger()
    if verbose == 1:
        logger.setLevel(logging.INFO)
    elif verbose > 1:
        logger.setLevel(logging.DEBUG)

    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(
        subplot_kw=dict(projection=getattr(crs, projection)())
    )
    ax.add_feature(countries(scale=scale))

    fig.show()
    plt.pause(-1)


if __name__ == "__main__":
    main()



from cartopy.feature import NaturalEarthFeature

# Some convenient features with overridable decent default parameters.

# TODO: themes


def countries(**kwargs):
    params = {
        "category": "cultural",
        "name": "admin_0_countries",
        "scale": "10m",
        "edgecolor": "#524c50",
        "facecolor": "none",
        "alpha": 0.5,
        **kwargs,
    }
    return NaturalEarthFeature(**params)


def rivers(**kwargs):
    params = {
        "category": "physical",
        "name": "rivers_lake_centerlines",
        "scale": "10m",
        "edgecolor": "#226666",
        "facecolor": "none",
        "alpha": 0.2,
        **kwargs,
    }
    return NaturalEarthFeature(**params)


def lakes(**kwargs):
    params = {
        "category": "physical",
        "name": "lakes",
        "scale": "10m",
        "edgecolor": "#226666",
        "facecolor": "#226666",
        "alpha": 0.2,
        **kwargs,
    }
    return NaturalEarthFeature(**params)


def ocean(**kwargs):
    params = {
        "category": "physical",
        "name": "ocean",
        "scale": "10m",
        "edgecolor": "#226666",
        "facecolor": "#226666",
        "alpha": 0.2,
        **kwargs,
    }
    return NaturalEarthFeature(**params)



from typing import Any, Dict, Optional, Tuple, TypeVar, Union

from pyproj import Proj, Transformer
from shapely.geometry import LineString, Point, Polygon, mapping, shape
from shapely.geometry.base import BaseGeometry
from shapely.ops import transform

from ...core import GeoObject
from ...utils.descriptors import OrientedShape
from ...utils.geometry import reorient
from ...utils.mixins import HBoxMixin, HTMLAttrMixin, HTMLTitleMixin
from ..requests import GeoJSONType, JSONType


def to_geometry(elt: JSONType) -> BaseGeometry:
    """Builds a shapely geometry based on JSON representations.

    Parses the usual `out geom` format for geometries.

    >>> str(to_geometry({'type': 'node', 'lon': 1.5, 'lat': 43.6}))
    'POINT (1.5 43.6)'

    """
    if elt["type"] == "node":
        return Point(elt["lon"], elt["lat"])

    nodes = elt.get("nodes", None)
    if nodes is not None:
        assert isinstance(nodes, list)
    shape = LineString if nodes is None or nodes[0] != nodes[-1] else Polygon
    return reorient(shape(list((p["lon"], p["lat"]) for p in elt["geometry"])))


T = TypeVar("T", bound="NodeWayRelation")


class NodeWayRelation(GeoObject, HBoxMixin, HTMLTitleMixin, HTMLAttrMixin):
    """Common behaviour of Node, Way and Relation classes.

    This class is de facto an abstract class. Instantiating it will always
    fallback to a child class registered in `subclasses`, based on the type
    of the feature (`json["type_"]`).

    - Representation for such objects in implemented with usual mixins.
    - Default simplification is available as well.

    """

    shape = OrientedShape()
    subclasses: Dict[str, Any] = dict()
    instances: Dict[int, "NodeWayRelation"] = dict()

    def __new__(cls, json: Optional[GeoJSONType] = None):

        if json is None:
            return super().__new__(cls)

        # Singleton instances
        id_ = json["id_"]
        if id_ in NodeWayRelation.instances.keys():
            return NodeWayRelation.instances[id_]
        # Otherwise dispatch on subclasses
        type_ = NodeWayRelation.subclasses[json["type_"]]
        if cls in NodeWayRelation.subclasses.values():
            return super().__new__(cls)
        else:
            return type_.__new__(type_, json)

    def __init__(self, json: GeoJSONType):
        super().__init__()
        self.json = json

    @classmethod
    def __init_subclass__(cls):
        super().__init_subclass__()
        NodeWayRelation.subclasses[cls.__name__.lower()] = cls

    @property
    def __geo_interface__(self) -> GeoJSONType:
        return mapping(self.shape)

    @property
    def html_attr_list(self):
        return sorted(
            key
            for key, value in self.json.items()
            if key not in ["geometry", "nodes", "members", "_parent"]
            and (
                key == "name:en"
                or not (key.startswith("name:") or "_name" in key)
            )
            and "wiki" not in key
            and not key.startswith("source:")
            and value == value
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__} {self.json}"

    def _repr_html_(self) -> str:
        return (
            super()._repr_html_()
            + "<div style='float: left; margin: 10px;'>"
            + self._repr_svg_()
            + "</div>"
        )

    def __getattr__(self, name):

        if name in ["__getstate__", "__setstate__"]:  # security for pickling
            raise AttributeError(name)

        if name.endswith("_"):  # in case the name is reserved
            name = name[:-1]

        value = self.json.get(name, None)
        if value is None:
            address = self.json.get("address", None)
            if address is not None:
                value = address.get(name, None)
        if value is None:
            raise AttributeError(name)
        return value

    def simplify(
        self: T,
        resolution: float,
        bounds: Union[None, Tuple[float, float, float, float]] = None,
    ) -> T:

        if bounds is None:
            bounds = self.parent.bounds

        proj = Proj(
            proj="aea",  # equivalent projection
            lat_1=bounds[1],
            lat_2=bounds[3],
            lat_0=(bounds[1] + bounds[3]) / 2,
            lon_0=(bounds[0] + bounds[2]) / 2,
        )

        forward = Transformer.from_proj(Proj("EPSG:4326"), proj, always_xy=True)
        backward = Transformer.from_proj(
            proj, Proj("EPSG:4326"), always_xy=True
        )
        new = type(self)(self.json)
        new.shape = transform(
            backward.transform,
            transform(forward.transform, self.shape).simplify(resolution),
        )
        return new


class Node(NodeWayRelation):
    """Node just delegates the shape parsing to shapely."""

    def __init__(self, json: GeoJSONType):
        super().__init__(json)
        self.shape = shape(self.json["geometry"])


class Way(NodeWayRelation):
    """Way just delegates the shape parsing to shapely."""

    def __init__(self, json: GeoJSONType):
        super().__init__(json)
        self.shape = shape(self.json["geometry"])


class Relation(NodeWayRelation):
    """Common behaviour for all Relation specifications.

    This class is de facto an abstract class. Instantiating it will always
    fallback to a child class registered in `subclasses`, based on the type
    of the feature (`json["type"]`).

    - json['type_'] is node, way, or relation
    - json['type'] is the `type` key in the `tags` subdictionary
    """

    subclasses: Dict[str, Any] = dict()

    def __new__(cls, json: Optional[GeoJSONType] = None):
        if json is None:
            return super().__new__(cls, json)
        if json["type"] not in Relation.subclasses:
            msg = (
                f"The parser for {json['type'].title()} is not implemented yet."
                f"\nIf you feel enthusiastic, you may implement a class "
                f"{json['type'].title()} which inherits from Relation"
                "\nand knows how to parse this relation."
            )
            raise NotImplementedError(msg)
        type_ = Relation.subclasses[json["type"]]
        if cls in Relation.subclasses.values():
            return super().__new__(cls, json)
        else:
            return type_.__new__(type_, json)

    @classmethod
    def __init_subclass__(cls):
        super().__init_subclass__()
        Relation.subclasses[cls.__name__.lower()] = cls



# fmt: off

import inspect
import logging
import pprint
from abc import ABCMeta
from typing import Any, Dict, List

from cartopy.crs import (
    Globe,
    InterruptedGoodeHomolosine,
    LambertCylindrical,
    Miller,
    Projection,
    Robinson,
    RotatedPole,
    Sinusoidal,
)

from pyproj import CRS, Proj, Transformer
from shapely import geometry
from shapely.ops import transform

from .vega_params import (  # These ones can be unpacked in altair!
    UTM,
    AlbersEqualArea,
    AzimuthalEquidistant,
    EqualEarth,
    EquidistantConic,
    Gnomonic,
    LambertAzimuthalEqualArea,
    LambertConformal,
    Mercator,
    Mollweide,
    Orthographic,
    Stereographic,
    TransverseMercator,
)

# fmt:on

_log = logging.getLogger(__name__)

base_classes: Dict["str", Projection] = {
    "lcc": LambertConformal,
    "laea": LambertAzimuthalEqualArea,
    "aea": AlbersEqualArea,
    "tmerc": TransverseMercator,
    "utm": UTM,
    "merc": Mercator,
    "cea": LambertCylindrical,
    "mill": Miller,
    "ob_tran": RotatedPole,  # or RotatedGeometric...
    "gnom": Gnomonic,
    "stere": Stereographic,
    "ortho": Orthographic,
    "eqearth": EqualEarth,
    "moll": Mollweide,
    "robin": Robinson,
    "igh": InterruptedGoodeHomolosine,
    "aeqd": AzimuthalEquidistant,
    "sinu": Sinusoidal,
    "eqdc": EquidistantConic,
}

arg_codes = {
    8801: "central_latitude",
    8802: "central_longitude",
    8803: "standard_parallel_1",
    8804: "standard_parallel_2",
    8805: "scale_factor",
    8806: "false_easting",
    8807: "false_northing",
    8821: "central_latitude",
    8822: "central_longitude",
    8823: "standard_parallel_1",
    8824: "standard_parallel_2",
    8825: "scale_factor",
    8826: "false_easting",
    8827: "false_northing",
}


class EPSGProjectionMeta(ABCMeta):
    @staticmethod
    def init_args(parameters: List[Dict[str, Any]], proj: Projection):
        all_ = dict(
            (arg_codes[elt["id"]["code"]], elt["value"]) for elt in parameters
        )
        if "standard_parallel_1" in all_ and "standard_parallel_2" in all_:
            all_["standard_parallels"] = (
                all_.pop("standard_parallel_1"),
                all_.pop("standard_parallel_2"),
            )
        # Get the arguments for the selected CRS projection
        init_args = inspect.getargs(proj.__init__.__code__).args
        remove_args = list(
            key for key, _value in all_.items() if key not in init_args
        )

        for key in remove_args:
            _log.warning(f"argument '{key}' ignored")
            del all_[key]

        return all_

    @staticmethod
    def projected_shape(bbox: Dict[str, float], transformer):
        y0 = bbox["south_latitude"]
        x0 = bbox["east_longitude"]
        y1 = bbox["north_latitude"]
        x1 = bbox["west_longitude"]

        return transform(
            transformer.transform,
            geometry.LineString(
                [(x0, y0), (x0, y1), (x1, y1), (x1, y0), (x0, y0)]
            ),
        )

    def __new__(metacls, name, bases, attr_dict):
        identifier = attr_dict.get("identifier", None)
        if identifier is None:
            for base in bases:
                identifier = vars(base).get("identifier", None)
                if identifier is not None:
                    break
            else:  # not found
                raise TypeError("No identifier in class or parent classes.")

        crs = CRS(identifier)
        basic_dict = crs.to_dict()
        _log.debug(pprint.pformat(basic_dict))
        base_class = base_classes.get(basic_dict["proj"], Projection)
        _log.debug(f"Generate a class based on {base_class.__name__}")

        transformer = Transformer.from_proj(
            Proj("epsg:4326"), Proj(crs), always_xy=True
        )
        attr_dict = {**crs.to_json_dict(), **attr_dict}
        _log.debug(f"Based on {pprint.pformat(attr_dict)}")

        if base_class is Projection:
            attr_dict["threshold"] = 1e4

        def __init__(self):
            if base_class is Projection:
                return Projection.__init__(self, crs.to_dict())

            ellipse = basic_dict.get("ellps", None)
            if ellipse is not None:
                globe = Globe(ellipse=ellipse)
            else:
                a, rf = basic_dict.get("a", None), basic_dict.get("rf", None)
                if a is None or rf is None:
                    msg = f"Fallback to regular Projection with {crs.to_dict()}"
                    _log.warning(msg)
                    return Projection.__init__(self, crs.to_dict())
                globe = Globe(semimajor_axis=a, inverse_flattening=rf)

            base_class.__init__(
                self,
                **EPSGProjectionMeta.init_args(
                    attr_dict["conversion"]["parameters"], base_class
                ),
                globe=globe,
            )

        attr_dict["__init__"] = __init__

        projected_shape = EPSGProjectionMeta.projected_shape(
            attr_dict["bbox"], transformer
        )

        def x_limits(self):
            x0, y0, x1, y1 = projected_shape.bounds
            return x0, x1

        attr_dict["x_limits"] = property(x_limits)

        def y_limits(self):
            x0, y0, x1, y1 = projected_shape.bounds
            return y0, y1

        attr_dict["y_limits"] = property(y_limits)

        def boundary(self):
            x0, x1 = self.x_limits
            y0, y1 = self.y_limits
            return geometry.LineString(
                [(x0, y0), (x0, y1), (x1, y1), (x1, y0), (x0, y0)]
            )

        attr_dict["boundary"] = property(boundary)

        attr_dict["projected_shape"] = projected_shape

        return super().__new__(metacls, name, (base_class,), attr_dict)


def __getattr__(name: str) -> Projection:
    if name.startswith("EPSG_"):
        return EPSGProjectionMeta(name, (), dict(identifier=f"EPSG:{name[5:]}"))

    raise AttributeError



from typing import Union

import pandas as pd
import pyproj
from shapely.geometry import shape

from ..core import GeoObject
from ..osm import Nominatim


def valid_crs(area: Union[str, GeoObject]) -> pd.DataFrame:
    """Returns a list of valid CRS according to a GeoObject."""
    if isinstance(area, str):
        candidate = Nominatim.search(area)
        if candidate is None:
            raise ValueError("No such place found.")
        area = candidate
    msg = "The argument must implement the __geo_interface__ protocol"
    if not isinstance(area, GeoObject):
        raise TypeError(msg)
    features = [
        "auth_name",  # EPSG
        "code",  # This one is of interest!
        "deprecated",  # mostly False
        "name",
        "projection_method_name",  # projection family
    ]
    return pd.DataFrame.from_records(
        list(
            {
                "area": crsinfo.area_of_use.name  # type: ignore
                # typing: explain (the hasattr catches it)
                if hasattr(crsinfo, "area_of_use") else None,
                **dict((col, getattr(crsinfo, col)) for col in features),
            }
            for crsinfo in pyproj.database.query_crs_info(
                "EPSG",
                "PROJECTED_CRS",  # type: ignore
                # typing: explain (str are also valid)
                pyproj.aoi.AreaOfInterest(
                    *shape(area.__geo_interface__).bounds
                ),
            )
        )
    )



from pathlib import Path
from typing import Generic, TypeVar, Union

from shapely.geometry.base import BaseGeometry

from .geometry import reorient

T = TypeVar("T")


class Descriptor(Generic[T]):
    def __set_name__(self, obj, name: str) -> None:
        self.public_name = name
        self.private_name = "_" + name

    def __get__(self, obj, cls=None) -> T:
        return getattr(obj, self.private_name)

    def __set__(self, obj, value: T) -> None:
        setattr(obj, self.private_name, value)


class DirectoryCreateIfNotExists(Descriptor[Path]):
    def __set__(self, obj, path: Union[str, Path]) -> None:
        if isinstance(path, str):
            path = Path(path)
        if path.exists() and not path.is_dir():
            raise RuntimeError(f"{path} exists and is not a directory")
        if not path.exists():
            path.mkdir(parents=True)
        setattr(obj, self.private_name, path)


class OrientedShape(Descriptor[BaseGeometry]):
    """Ensures the given shape is properly oriented for Altair.

    Reference: https://altair-viz.github.io/user_guide/data.html?highlight=orient%20transform#winding-order
    """

    def __set__(self, obj, shape: BaseGeometry):
        setattr(obj, self.private_name, reorient(shape, orientation=-1))



import warnings

import importlib_metadata
from tqdm import TqdmExperimentalWarning

# Monkey-patches GeoAxesSubplot for .set_extent()
# Monkey-patches GeoDataFrame for .extent()
from .utils import geoaxes  # noqa: F401

__version__ = importlib_metadata.version("cartes")


# Silence this warning about autonotebook mode for tqdm
warnings.simplefilter("ignore", TqdmExperimentalWarning)



from cartes.osm import Nominatim


def test_search() -> None:
    toulouse = Nominatim.search("Toulouse")
    assert toulouse is not None
    assert toulouse.display_name.startswith("Toulouse")
    assert toulouse.city == "Toulouse"
    assert toulouse.osm_id == 35738

    assert "2154" in list(toulouse.valid_crs().code)


def test_lookup() -> None:
    capitole = Nominatim.lookup("R367073")
    assert capitole is not None
    assert capitole.city == "Toulouse"  # type: ignore
    assert capitole.category == "tourism"  # type: ignore


def test_reverse() -> None:
    musee = Nominatim.reverse(43.608, 1.442)
    assert musee is not None
    assert musee.name == "Basilique Saint-Sernin"
    assert musee.road == "Place Saint-Sernin"
    assert musee.osm_type == "way"



import pytest

from cartes.osm.overpass import Overpass


def test_basic_query() -> None:
    query_lfbo = "[out:json];area[icao=LFBO];nwr(area)[aeroway];out geom;"
    lfbo = Overpass.request(query=query_lfbo)
    runways = sorted(set(lfbo.data.query('aeroway=="runway"').ref))
    assert runways == ["14H/32H", "14L/32R", "14R/32L"]


def test_proper_query() -> None:
    lfbo = Overpass.request(area=dict(icao="LFBO"), aeroway=True)
    runways = sorted(set(lfbo.data.query('aeroway=="runway"').ref))
    assert runways == ["14H/32H", "14L/32R", "14R/32L"]


def test_validator() -> None:
    query_mercantour = (
        "[out:json][timeout:180];"
        "area['boundary'='protected_area']['name'~'Mercantour']->.a;"
        "rel(pivot.a);"
        "out geom;"
        "node[natural=peak](area.a);"
        "out geom;"
    )
    mercantour = Overpass.request(query=query_mercantour)
    with pytest.raises(TypeError):
        mercantour.drop(columns=["geometry"])


def test_generate_query() -> None:
    assert (
        Overpass.build_query(area=dict(icao="LFBO"), aeroway=True)
        == "[out:json][timeout:180];"
        "area[icao=LFBO];nwr(area)[aeroway];out geom;"
    )

    south, west, north, east = 33, -30, 70, 35
    assert (
        Overpass.build_query(
            bounds=[west, south, east, north], node=dict(capital="yes")
        )
        == "[out:json][timeout:180][bbox:33,-30,70,35];"
        "node[capital=yes];out geom;"
    )

    mercantour = (
        "[out:json][timeout:180];"
        'area[boundary=protected_area][name~"Mercantour"]->.a;'
        "rel(pivot.a);"
        "out geom;"
        "node(area.a)[natural=peak];"
        "out geom;"
    )
    assert (
        Overpass.build_query(
            area=dict(
                boundary="protected_area",
                name=dict(regex="Mercantour"),
                as_="a",
            ),
            nwr=[
                dict(rel=dict(area="a")),
                dict(node=dict(natural="peak", area="a")),
            ],
        )
        == mercantour
    )

    toulouse = (
        "[out:json][timeout:180];"
        "area[name=Toulouse][admin_level=8];"
        "rel(area)[boundary=postal_code];out geom;"
    )

    assert (
        Overpass.build_query(
            area=dict(name="Toulouse", admin_level=8),
            rel=dict(boundary="postal_code"),
        )
        == toulouse
    )

    cannes = (
        "[out:json][timeout:180];"
        "area[name=Cannes][admin_level=8];"
        "nwr(area)[building];"
        "out geom;"
    )

    assert (
        Overpass.build_query(
            area=dict(name="Cannes", admin_level=8), building=True
        )
        == cannes
    )

    neste = (
        "[out:json][timeout:180];"
        'rel[waterway=canal][name~"Neste"];'
        "out geom;"
        "rel(around:1000)[waterway=river];"
        "out geom;"
        'rel[waterway=river][name="La Garonne"];'
        "out geom;"
    )

    assert (
        Overpass.build_query(
            rel=[
                dict(waterway="canal", name=dict(regex="Neste")),
                dict(around=1000, waterway="river"),
                dict(waterway="river", name="La Garonne"),
            ]
        )
        == neste
    )



import logging
from pathlib import Path

from cartes.osm.requests import json_request


def pytest_configure(config):
    _log = logging.getLogger()
    _log.setLevel(logging.INFO)

    json_request.cache_dir = Path(config.rootdir) / "cartes" / "tests" / "cache"
    _log.warning(f"Using cache_dir {json_request.cache_dir} for tests")



from __future__ import annotations

from abc import ABC, abstractproperty
from typing import TYPE_CHECKING, Any, Iterator, Tuple

from cartopy.crs import Orthographic

import pandas as pd
from pyproj import Proj, Transformer
from shapely.geometry import base, shape
from shapely.ops import transform

if TYPE_CHECKING:
    from ipyleaflet import Map, Marker, Polygon, Polyline


class GeoObject(ABC):
    @abstractproperty
    def __geo_interface__(self):
        return 0

    @classmethod
    def __subclasshook__(cls, C):
        if cls is GeoObject:
            if any("__geo_interface__" in B.__dict__ for B in C.__mro__):
                return True
        return NotImplemented

    @property
    def shape(self) -> base.BaseGeometry:
        return shape(self.__geo_interface__)

    @property
    def ortho_shape(self):
        shape_ = shape(self)
        lon, lat = shape_.centroid.coords[0]
        proj = Orthographic(lon, lat)
        transformer = Transformer.from_proj(
            Proj("EPSG:4326"), Proj(proj.proj4_init), always_xy=True
        )
        return transform(transformer.transform, shape_)

    @property
    def equivalent_shape(self):
        shape_ = shape(self)
        bounds = self.bounds
        proj = Proj(
            proj="aea",  # equivalent projection
            lat_1=bounds[1],
            lat_2=bounds[3],
            lat_0=(bounds[1] + bounds[3]) / 2,
            lon_0=(bounds[0] + bounds[2]) / 2,
        )

        transformer = Transformer.from_proj(
            Proj("EPSG:4326"), proj, always_xy=True
        )
        return transform(transformer.transform, shape_)

    @property
    def area(self) -> float:
        return self.equivalent_shape.area

    @property
    def bounds(self) -> Tuple[float, float, float, float]:
        return self.shape.bounds

    @property
    def extent(self) -> Tuple[float, float, float, float]:
        west, south, east, north = self.bounds
        return west, east, south, north

    @property
    def _geom(self):
        # convenient for cascaded_union
        return self.shape._geom

    @property
    def type(self):
        # convenient for intersections, etc.
        return self.shape.type

    def _repr_svg_(self):
        return self.ortho_shape._repr_svg_()

    def valid_crs(self) -> pd.DataFrame:
        from .crs.info import valid_crs

        return valid_crs(self)

    def zoom_level(self) -> int:
        x1, y1, x2, y2 = self.bounds

        if x1 == x2:
            return 17

        def func(n):
            factor = (360 * 2**-n) / (x2 - x1)
            if factor < 1:
                return 1 / factor
            return factor

        return min(range(20), key=func)

    def leaflet(
        self, **kwargs: Any
    ) -> None | "Polygon" | "Polyline" | "Marker":
        """Returns a Leaflet layer to be directly added to a Map.

        .. warning::
            This is only available if the Leaflet `plugin <plugins.html>`_ is
            activated. (true by default)

        The elements passed as kwargs as passed as is to the Polygon __init__
        method.
        """
        from ipyleaflet import Marker, Polygon, Polyline

        kwargs = {**dict(weight=3), **kwargs}
        coords: list[Any] = []

        def unfold(shape: Any) -> Iterator[Any]:
            yield shape.exterior
            yield from shape.interiors

        if self.shape.geom_type == "Polygon":
            coords = list(
                list((lat, lon) for (lon, lat, *_) in x.coords)
                for x in unfold(self.shape)
            )
            return Polygon(locations=coords, **kwargs)
        if self.shape.geom_type == "MultiPolygon":
            coords = list(
                list((lat, lon) for (lon, lat, *_) in x.coords)
                for piece in self.shape.geoms
                for x in unfold(piece)
            )
            return Polygon(locations=coords, **kwargs)
        if self.shape.geom_type == "LineString":
            return Polyline(
                locations=list((lat, lon) for (lon, lat) in self.shape.coords),
                **kwargs,
            )
        if self.shape.geom_type == "MultiLineString":
            return Polyline(
                locations=list(
                    list((lat, lon) for (lon, lat) in piece.coords)
                    for piece in self.shape.geoms
                ),
                **kwargs,
            )
        if self.shape.geom_type == "Point":
            (lon, lat), *_ = self.shape.coords
            return Marker(location=(lat, lon), **kwargs)

        return None

    def map_leaflet(self, **kwargs) -> "Map":
        from ipyleaflet import Map

        x1, y1, x2, y2 = self.bounds
        m = Map(
            center=(0.5 * (y1 + y2), 0.5 * (x1 + x2)), zoom=self.zoom_level()
        )

        layer = self.leaflet(**kwargs)
        if layer is not None:
            m.add_layer(layer)

        return m



from __future__ import annotations

import json
import logging
from functools import cached_property  # noqa: F401
from pathlib import Path
from typing import Any, Callable, Dict, Generic, Hashable, TypeVar, Union

import pandas as pd

from .descriptors import DirectoryCreateIfNotExists

T = TypeVar("T")

_log = logging.getLogger(__name__)


class CacheFunction(Generic[T]):
    """Function caching utilities.

    This class puts results in a two-layer cache.

    Results are first fetched from a LRU cache available as long as the
    function exists. If the result is not in cache, then it attempts picking
    it from a file in a dedicated directory.

    The class needs four arguments:
    - cache_dir: Union[str, Path] is the place where to store the cache files;
    - hashing: Callable[..., Hashable] computes a hash based on arguments
      passed to the function. The hash will be the key to the LRU cache and the
      name of the file;
    - writer: Callable[[T, Path], None] describes how to serialize the results;
    - reader: Callable[[Path], T] describes how to read the results from a file.

    """

    cache_dir = DirectoryCreateIfNotExists()

    def __init__(
        self,
        function: Callable[..., T],
        cache_dir: Union[str, Path],
        hashing: Callable[..., Hashable],
        writer: Callable[[T, Path], None],
        reader: Callable[[Path], T],
    ):
        self.function = function
        self.cache_dir = cache_dir
        self.hashing = hashing
        self.writer = writer
        self.reader = reader

        self.lru_cache: Dict[Hashable, T] = dict()

    def __call__(self, *args, **kwargs) -> T:
        hashcode = self.hashing(*args, **kwargs)
        cache_file = self.cache_dir / str(hashcode)
        res = self.lru_cache.get(cache_file.as_posix(), None)  # type: ignore

        if res is not None:
            return res

        _log.info(f"Looking for cache file: {cache_file}")

        if cache_file.exists():
            _log.info(f"Using cache file: {cache_file}")
            res = self.reader(cache_file)

        if res is None:
            msg = f"Calling function {self.function} with {args, kwargs}"
            _log.debug(msg)
            res = self.function(*args, **kwargs)
            self.writer(res, cache_file)

        self.lru_cache[cache_file.as_posix()] = res
        return res


class CacheResults(Generic[T]):
    """Defines a decorator for functions when results should be cached.

    The class needs four arguments:
    - cache_dir: Union[str, Path] is the place where to store the cache files;
    - hashing: Callable[..., Hashable] computes a hash based on arguments
      passed to the function. The hash will be the key to the LRU cache and the
      name of the file;
    - writer: Callable[[T, Path], None] describes how to serialize the results;
    - reader: Callable[[Path], T] describes how to read the results from a file.

    """

    def __init__(
        self,
        cache_dir: Union[str, Path],
        hashing: Callable[..., Hashable],
        writer: Callable[[T, Path], None],
        reader: Callable[[Path], T],
    ):
        self.cache_dir = cache_dir
        self.hashing = hashing
        self.writer = writer
        self.reader = reader

    def __call__(self, function: Callable[..., T]) -> CacheFunction[T]:
        cache_function = CacheFunction(
            function, self.cache_dir, self.hashing, self.writer, self.reader
        )
        cache_function.__doc__ = function.__doc__
        return cache_function


def write_json(content: Any, cache_file: Path) -> None:
    _log.info(f"Writing cache file {cache_file}")
    directory = cache_file.parent
    if not directory.exists():
        directory.mkdir(parents=True)
    cache_file.write_text(json.dumps(content))


def read_json(cache_file: Path) -> Any:
    _log.info(f"Reading cache file {cache_file}")
    return json.loads(cache_file.read_text())


def write_df_json(df: pd.DataFrame, cache_file: Path) -> None:
    _log.info(f"Writing cache file {cache_file}")
    directory = cache_file.parent
    if not directory.exists():
        directory.mkdir(parents=True)
    df.to_json(cache_file, orient="records")


def read_json_df(cache_file: Path) -> pd.DataFrame | None:
    _log.info(f"Reading cache file {cache_file}")
    return pd.read_json(cache_file)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cartes',
 'cartes.atlas',
 'cartes.crs',
 'cartes.dataviz',
 'cartes.dataviz.markers',
 'cartes.osm',
 'cartes.osm.overpass',
 'cartes.osm.overpass.relations',
 'cartes.tests',
 'cartes.utils']

package_data = \
{'': ['*'], 'cartes.tests': ['cache/*']}

install_requires = \
['Cartopy>=0.19',
 'Shapely>=1.8,<2.0',
 'aiohttp>=3.8,<4.0',
 'altair>=4.2,<5.0',
 'appdirs>=1.4,<2.0',
 'beautifulsoup4>=4.10,<5.0',
 'geopandas>=0.10',
 'importlib-metadata>=4.2,<5.0',
 'jsonschema>=3.0,<4.0',
 'lxml>=4.7,<5.0',
 'matplotlib>=3.5,<4.0',
 'numpy>=1.21',
 'pandas>=1.3,<2.0',
 'pyproj>=3.1',
 'requests>=2.27,<3.0',
 'scipy>=1.7,<2.0',
 'tqdm>=4.62,<5.0']

entry_points = \
{'console_scripts': ['cartes = cartes.__main__:main']}

setup_kwargs = {
    'name': 'cartes',
    'version': '0.7.3',
    'description': 'A generic toolbox for building maps in Python',
    'long_description': '# Cartes\n\n![build](https://github.com/xoolive/cartes/workflows/build/badge.svg)\n![docs](https://github.com/xoolive/cartes/actions/workflows/github-pages.yml/badge.svg)\n[![Code Coverage](https://img.shields.io/codecov/c/github/xoolive/cartes.svg)](https://codecov.io/gh/xoolive/cartes)\n[![Checked with mypy](https://img.shields.io/badge/mypy-checked-blue.svg)](https://mypy.readthedocs.io/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-black.svg)](https://github.com/psf/black)\n![License](https://img.shields.io/pypi/l/cartes.svg)\\\n![PyPI version](https://img.shields.io/pypi/v/cartes)\n[![PyPI downloads](https://img.shields.io/pypi/dm/cartes)](https://pypi.org/project/cartes)\n![Conda version](https://img.shields.io/conda/vn/conda-forge/cartes)\n[![Conda Downloads](https://img.shields.io/conda/dn/conda-forge/cartes.svg)](https://anaconda.org/conda-forge/cartes)\n\nCartes is a Python library providing facilities to produce meaningful maps.\n\nCartes builds on top of most common Python visualisation libraries (Matplotlib/Cartopy, Altair, ipyleaflet) and data manipulation libraries (Pandas, Geopandas) and provides mostly:\n\n- a **comprehensive set of geographic projections**, built on top of Cartopy and Altair/d3.js;\n- an **interface to OpenstreetMap Nominatim and Overpass API**. Result of requests are parsed in a convenient format for preprocessing and storing in standard formats;\n- beautiful **default parameters** for quality visualisations;\n- **advanced caching facilities**. Do not download twice the same content in the same day.\n\nThe cartes library is a powerful asset to **publish clean, lightweight geographical datasets**; and to **produce decent geographical visualisations** in few lines of code.\n\n## Gallery\n\n<a href="https://cartes-viz.github.io/gallery/mercantour.html"><img width="20%" src="https://cartes-viz.github.io/_static/homepage/mercantour.png"></a>\n<a href="https://cartes-viz.github.io/gallery/footprint.html"><img width="20%" src="https://cartes-viz.github.io/_static/homepage/antibes.png"></a>\n<a href="https://cartes-viz.github.io/gallery/airports.html"><img width="20%" src="https://cartes-viz.github.io/_static/homepage/airports.png"></a>\n<a href="https://cartes-viz.github.io/gallery/tokyo_metro.html#zoom-in-to-downtown-tokyo"><img width="20%" src="https://cartes-viz.github.io/_static/homepage/tokyo.png"></a>\n\nMore in the [documentation](https://cartes-viz.github.io/gallery.html)\n\n## Installation\n\nLatest release:\n\nRecommended, with conda:\n```sh \nconda install -c conda-forge cartes\n```\n\nor with pip:\n```sh\npip install cartes\n```\n\nDevelopment version:\n\n```sh\ngit clone https://github.com/xoolive/cartes\ncd cartes\npip install .\n```\n\n## Documentation\n\n![docs](https://github.com/xoolive/cartes/actions/workflows/github-pages.yml/badge.svg)\n\nDocumentation available at https://cartes-viz.github.io/\n',
    'author': 'Xavier Olive',
    'author_email': 'git@xoolive.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from .topo import GithubAPI, NpmAPI, TopoCatalogue

__all__ = ["default", "world_atlas"]

default = TopoCatalogue(
    api=GithubAPI(username="deldersveld", repository="topojson")
)

world_atlas = TopoCatalogue(api=NpmAPI(name="world-atlas@2.0.2"))


def __getattr__(name: str):
    return getattr(default, name)



from cartes.osm import Overpass


def test_multipolygon():
    airport = Overpass.request(
        query="""[out:json][timeout:180];area[icao="LFPO"];nwr(area)[aeroway];
        out geom;""",
    )
    assert airport.data.shape[0] == 361

    airport = Overpass.request(
        query="""[out:json][timeout:180];area[icao="LEBL"];nwr(area)[aeroway];
        out geom;""",
    )
    assert airport.data.shape[0] == 864

    airport = Overpass.request(
        query="""[out:json][timeout:180];area[icao="EDDF"];nwr(area)[aeroway];
        out geom;""",
    )
    assert airport.data.shape[0] == 5297


def test_simplify():
    tls = Overpass.request(
        query="""[out:json][timeout:180];
area[name='Toulouse'][admin_level=8];
rel(area)["boundary"="postal_code"];
out geom;
 """
    )
    # Useless test, just check it works
    assert tls.simplify(1e3, max_workers=1).data.shape[0] == 6

    assert tls.simplify(1e3).data.shape[0] == 6



from __future__ import annotations

import geopandas as gpd
from cartopy.crs import Projection
from cartopy.img_transform import mesh_projection
from cartopy.mpl.geoaxes import GeoAxesSubplot

import numpy as np
from pyproj import Proj, Transformer
from shapely.geometry import Polygon, box
from shapely.geometry.base import BaseGeometry
from shapely.ops import transform

from ..core import GeoObject
from ..osm import Nominatim
from ..utils.geometry import fix_geodataframe

# We patch the set_extent method to use GeoObjects instead.


def _set_extent(self, shape, buffer: float = 0.01):
    if isinstance(shape, str):
        shape = Nominatim.search(shape)
    if isinstance(shape, GeoObject):
        x1, x2, y1, y2 = shape.extent
        extent = (x1 - buffer, x2 + buffer, y1 - buffer, y2 + buffer)
        return self._set_extent(extent)
    self._set_extent(shape)


GeoAxesSubplot._set_extent = GeoAxesSubplot.set_extent
GeoAxesSubplot.set_extent = _set_extent


def make_polygon(projection, x1, x2, y1, y2):
    X, Y, _extent = mesh_projection(projection, 100, 100, (x1, x2), (y1, y2))
    x = np.r_[X[:, 0], X[-1, 1:], X[-2::-1, -1], X[0, -2::-1]].tolist()
    y = np.r_[Y[:, 0], Y[-1, 1:], Y[-2::-1, -1], Y[0, -2::-1]].tolist()
    return Polygon(list(zip(x, y)))


def gpd_extent(
    gdf: gpd.GeoDataFrame,
    shape,
    projection: Projection | None = None,  # noqa: B008
    buffer: float = 0.01,
) -> BaseGeometry:
    """Computes the intersection of all geometries in a bounding box.

    1. Creates a square boundingbox in the destination CRS (projection);
    2. Filters out geometries not intersecting the bounding box;
    3. Tentatively fix invalid geometries;
    4. Computes intersection with the projected bounding box

    """

    if isinstance(shape, str):
        shape = Nominatim.search(shape)
    if isinstance(shape, GeoObject):
        x1, x2, y1, y2 = shape.extent
        x1, x2, y1, y2 = (x1 - buffer, x2 + buffer, y1 - buffer, y2 + buffer)
    else:
        x1, x2, y1, y2 = shape

    if projection is not None:
        proj = Proj(projection.proj4_init)

        box_extent = make_polygon(projection, x1, x2, y1, y2)
        transformer = Transformer.from_proj(
            Proj("EPSG:4326"), proj, always_xy=True
        )
        projected_box = transform(transformer.transform, box_extent)

        x1, y1, x2, y2 = projected_box.bounds
        square_in_projected = make_polygon(projection, x1, x2, y1, y2)
        transformer = Transformer.from_proj(
            proj, Proj("EPSG:4326"), always_xy=True
        )
        shape_in_latlon = transform(transformer.transform, square_in_projected)
    else:
        shape_in_latlon = box(x1, y1, x2, y2)

    return (
        gdf.loc[gdf.geometry.intersects(shape_in_latlon)]
        .pipe(fix_geodataframe)
        .assign(geometry=lambda df: df.geometry.intersection(shape_in_latlon))
    )


gpd.GeoDataFrame.extent = gpd_extent



import json
from pathlib import Path

from matplotlib import path

import numpy as np

current_dir = Path(__file__).parent
__all__ = list(p.stem for p in current_dir.glob("*.json"))


def __getattr__(name: str) -> path.Path:
    file_path = current_dir / (name + ".json")
    if file_path.exists():
        data = json.loads(file_path.read_text())
        return path.Path(
            vertices=data["vertices"], codes=np.array(data["codes"], np.uint8)
        )

    raise AttributeError(
        f"No {name}.json file found in {current_dir.absolute()}."
    )
